# Author: Hansheng Zhao <copyrighthero@gmail.com> (https://www.zhs.me)

from .List import List


__all__ = ('Queue', )


class Queue(List):
  """ Blocking queue based on redis list """

  __slots__ = ()

  def __init__(self, redis, token = None):
    """
    Queue interface constructor
    :param redis: redis, the instance
    :param token: str|bytes, the token
    """
    super(Queue, self).__init__(redis, token)

  @staticmethod
  def _timeout(timeout):
    """
    Lint blocking timeout parameter
    :param timeout: int|float|mixed, the timeout
    :return: int, the linted timeout
    """
    # check if timeout is numeric
    timeout = int(timeout) \
      if isinstance(timeout, (int, float)) else 0
    # check if timeout is positive
    timeout = timeout if timeout >= 0 else 0
    return timeout

  @staticmethod
  def full():
    """
    Check if queue is full
    :return: False, always hungry
    """
    return False

  def empty(self):
    """
    Check if queue is empty
    :return: bool, whether empty
    """
    return self.length == 0

  def _get_from_multi(
    self,
    token=None, block=True, timeout=0, flag = True
  ):
    """
    Blocking get an item from the queue
    :param token: mixed, the queue token(s)
    :param block: bool, whether to block
    :param timeout: int, block length
    :param flag: bool, whether get right
    :return: mixed, an item
    """
    # acquire resources
    redis = self._redis
    convert = self._convert
    # sanitize token parameter
    token = tuple(map(lambda _: convert(_), token)) \
      if isinstance(
        token, (list, tuple, range, set, frozenset)
      ) else (
        None if token is None else (convert(token), )
      )
    token = (self._token, ) \
      if token is None else (self._token, *token)
    # lint timeout parameter
    timeout = self._timeout(timeout) if block else 0
    # check flag
    if block:
      # acquire from a list of lists
      result = redis.brpop(token, timeout) \
        if flag else redis.blpop(token, timeout)
      # deserialize result
      if result is not None:
        result = self._loads(result[1])
    else:
      result = None
      # acquire from a list of lists
      for item in token:
        result = redis.rpop(item) \
          if flag else redis.lpop(item)
        # deserialize result
        if result is not None:
          result = self._loads(result)
          break
    return result

  def _get(
    self, block = True, timeout = 0, flag = True
  ):
    """
    Blocking get an item from the queue
    :param block: bool, whether to block
    :param timeout: int, block length
    :param flag: bool, whether get right
    :return: mixed, an item
    """
    # acquire resources
    redis = self._redis
    token = self._token
    # sanitize timeout value
    timeout = self._timeout(timeout) if block else 0
    # check block flag
    if block:
      # acquire a result from queue
      result = redis.brpop((token, ), timeout) \
        if flag else redis.blpop((token, ), timeout)
      # deserialize result
      if result is not None:
        result = self._loads(result[1])
    else:
      # acquire non-blocking
      result = redis.rpop(token) \
        if flag else redis.lpop(token)
      # deserialize result
      if result is not None:
        result = self._loads(result)
    return result

  def _put(self, value, flag = True):
    """
    Non-blocking put an item to redis queue
    :param value: mixed, an item or message
    :param flag: bool, whether to put left
    :return: None
    """
    self.prepend(value) if flag else self.append(value)

  def get(self, block = True, timeout = 0):
    """
    Blocking get an item from the tail
    :param block: bool, whether to block
    :param timeout: int, block length
    :return: mixed, an item
    """
    return self._get(block, timeout, True)

  # aliases for get
  recv = get

  def get_left(self, block = True, timeout = 0):
    """
    Blocking get an item from the head
    :param block: bool, whether to block
    :param timeout: int, block length
    :return: mixed, an item
    """
    return self._get(block, timeout, False)

  # aliases for get_left
  recv_left = get_left

  def put(self, value):
    """
    Non-blocking put an item to redis queue
    :param value: mixed, an item or message
    :return: None
    """
    self._put(value, True)

  # aliases for put
  set = send = put

  def put_right(self, value):
    """
    Non-blocking put an item to redis queue
    :param value: mixed, an item or message
    :return: None
    """
    self._put(value, False)

  # aliases for put_right
  set_right = send_right = put_right

  def circulate(
    self, token = None, block = True, timeout = 0
  ):
    """
    Blocking circulate an item (BRPOPLPUSH)
    :param token: None|mixed, the target
    :param block: bool, whether block or not
    :param timeout: int, timeout length
    :return: mixed, the item
    """
    # acquire resources
    redis = self._redis
    # convert token to bytes
    token = self._convert(token) \
      if token is not None else self._token
    # lint timeout parameter
    timeout = self._timeout(timeout) if block else 0
    # acquire and circulate an item
    result = redis.brpoplpush(
      self._token, token, timeout
    ) if block else redis.rpoplpush(self._token, token)
    # deserialize result
    if result is not None: result = self._loads(result)
    return result

  # alias for get & pur message
  msg = property(get, put)



# Author: Hansheng Zhao <zhaohans@msu.edu> (https://www.zhs.me)

import bz2
import zlib
import json
import pickle

try:
  # attempts to import msgpack
  import msgpack
except ModuleNotFoundError:
  # use pickle if not available
  msgpack = pickle


__all__ = ('Serialize', )


class Serialize(object):
  """ Serialize class for data processing """

  __slots__ = (
    '_serialize', '_compress',
    '_serializer','_compressor'
  )

  def __init__(self, ser = None, com = None, **kwargs):
    """
    Initialize method for building object
    :param ser: str|dict|None, the serialize config
    :param com: str|None, the compress config
    :param kwargs: dict, the stray keyword args
    """
    # default configurations
    self._serialize = 'json'
    self._compress = 'zlib'
    self._serializer = None
    self._compressor = None
    # acquire serialize (and compress) config
    if ser is not None:
      if isinstance(ser, str):
        self._serialize = ser.lower()
      else:
        if 'serialize' in ser:
          self._serialize = ser['serialize']
        if 'compress' in ser:
          self._compress = ser['compress']
    elif 'serialize' in kwargs:
      self._serialize = kwargs['serialize']
    # acquire compress config
    if com is not None and isinstance(com, str):
      self._compress = com.lower()
    elif 'compress' in kwargs:
      self._compress = kwargs['compress']

  def __call__(self, payload, switch = True):
    """
    Alias for serialize/de-serialize methods
    :param payload: mixed, various objects
    :param switch: bool, whether to serialize
    :return: mixed, objects
    """
    return self.dumps(payload) \
      if switch else self.loads(payload)

  def _instantiate(self):
    """
    Instantiate serializer and compressor
    :return: None
    """
    # setup appropriate serializer
    if self._serializer is None:
      self._serializer = json
      if self._serialize == 'pickle':
        self._serializer = pickle
      elif self._serialize == 'msgpack':
        self._serializer = msgpack
    # setup appropriate compressor
    if self._compressor is None:
      self._compressor = zlib
      if self._compress == 'bz2':
        self._compressor = bz2

  # dumps method for python convention
  def dumps(self, payload):
    """
    Serialize and compress the payload
    :param payload: mixed, serializable object
    :return: byte, the serialized and compressed object
    """
    # instantiate serializer and compressor
    self._instantiate()
    # serialize payload
    payload = self._serializer.dumps(payload)
    # encode payload if payload is string
    if isinstance(payload, (str,)): payload = payload.encode()
    # compress and return payload
    return self._compressor.compress(payload)

  def dump(self, payload, file):
    """
    Serialize, compress and save
    :param payload: mixed, serializable object
    :param file: file descriptor
    :return: None
    """
    file.write(self.dumps(payload))
    file.flush()

  # alias for dump
  pack = dump

  # loads method for python convention
  def loads(self, payload):
    """
    Decompress and deserialize the payload
    :param payload: bytes, serialized byte string
    :return: object, the decompressed and deserialized object
    """
    # instantiate serializer and compressor
    self._instantiate()
    # decompress payload
    payload = self._compressor.decompress(payload)
    # deserialize and return payload
    return self._serializer.loads(payload, encoding='UTF8')

  def load(self, file):
    """
    Decompress and deserialize the file
    :param file: file descriptor, or BytesIO instance
    :return: object, the decompressed and deserialized object
    """
    # decompress, deserialize and return payload
    return self.loads(file.read())

  # alias for load
  unpack = load



# Author: Hansheng Zhao <copyrighthero@gmail.com> (https://www.zhs.me)

from .BaseStructure import BaseStructure


__all__ = ('Dict', )


class Dict(BaseStructure):
  """
  Dict interface for redis hashes

  !! NO AUTO WRITE-BACK SUPPORT !!
  !! BEHAVES LIKE SHELVE MODULE !!
  """

  __slots__ = ()

  def __init__(self, redis, token = None):
    """
    Dict interface constructor
    :param redis: redis, the instance
    :param token: mixed, the token
    """
    super(Dict, self).__init__(redis, token)
    # set the default data type
    self._type = b'hash'
    self._initiate()

  def __contains__(self, key):
    """
    Check if a field exists in the store
    :param key: mixed, serializable items
    :return: bool, whether item exists
    """
    return self._redis.hexists(
      self._token, self._convert(key)
    )

  def __delattr__(self, key):
    """
    Attempt to remove the attribute
    :param key: str, the attribute
    :return: None
    """
    try:
      # attempt to remote attribute first
      super(Dict, self).__delattr__(key)
    except AttributeError:
      # delete from database
      self.__delitem__(key)

  def __delitem__(self, key):
    """
    Remove a field from the storage
    :param key: mixed, the key field
    :return: bool, whether removed
    """
    return self._redis.hdel(
      self._token, self._convert(key)
    )

  def __eq__(self, other):
    """
    Check if items are the same
    :param other: dict|Dict, dict-like
    :return: bool, equal or not
    """
    if len(self) == len(other):
      return dict(self) == (
        dict(other)
        if isinstance(other, Dict)
        else other
      )
    else:
      return False

  def __getattr__(self, key):
    """
    Acquire an item from database
    :param key: mixed, the key field
    :return: mixed, the stored value
    """
    return self.__getitem__(key)

  def __getitem__(self, key):
    """
    Acquire an item from database
    :param key: mixed, the key field
    :return: mixed, the stored value
    """
    # acquire stored value
    result = self._redis.hget(
      self._token, self._convert(key)
    )
    # deserialize result
    if result is not None:
      result = self._loads(result)
    return result

  def __iter__(self):
    """
    Iterator to get keys in store
    :return: generator, the iterator
    """
    yield from self._redis.hkeys(self._token)

  def __len__(self):
    """
    Acquire key fields count
    :return: int, the count
    """
    return self._redis.hlen(self._token)

  def __ne__(self, other):
    """
    Check if two instances are same
    :param other: other instance
    :return: bool, whether not equal
    """
    return not self.__eq__(other)

  def __setattr__(self, key, value):
    """
    Attempt to set an attribute
    :param key: str, the attribute
    :param value: mixed, attribute
    :return: None
    """
    try:
      super(Dict, self).__setattr__(key, value)
    except AttributeError:
      self.__setitem__(key, value)

  def __setitem__(self, key, value):
    """
    Store a key-value pair to store
    :param key: mixed, the key field
    :param value: mixed, serializable
    :return: bool, whether new field
    """
    return self._redis.hset(
      self._token,
      self._convert(key),
      self._dumps(value)
    ) == 1

  # alias for set item
  set = __setitem__

  @property
  def length(self):
    """
    Property for key count
    :return: int, the count
    """
    return self.__len__()

  def copy(self):
    """
    Return a new instance
    :return: instance
    """
    instance = Dict(self._redis, self._token)
    instance.serialize = self._serialize
    return instance

  @staticmethod
  def fromkeys(*args, **kwargs):
    """
    Method for dict compatibility
    :param args: mixed, arguments
    :param kwargs: keyword arguments
    :return: dict, the built dict
    """
    return dict.fromkeys(*args, **kwargs)

  def get(self, key, default = None):
    """
    Acquire an item from store
    :param key: mixed, the key
    :param default: mixed, default
    :return: mixed, value/default
    """
    return self.__getitem__(key) \
      if key in self else default

  def items(self):
    """
    Acquire all items from data store
    :return: tuple<tuple>, the k-v pairs
    """
    yield from map(
      lambda _: (_[0], self._loads(_[1])),
      self._redis.hgetall(self._token).items()
    )

  def keys(self):
    """
    Acquire all keys in the dictionary
    :return: generator, the keys
    """
    yield from self._redis.hkeys(self._token)

  def pop(self, key):
    """
    Acquire item and delete key
    :param key: mixed, the key
    :return: mixed, stored value
    """
    # acquire item delete key
    result = self.__getitem__(key)
    self.__delitem__(key)
    return result

  def popitem(self):
    """
    Pop a pair of key & value
    :return: tuple, k-v pair
    """
    key = next(self.keys())
    value = self.pop(key)
    return key, value

  def setdefault(self, key, default = None):
    """
    Get or set the default into store
    :param key: mixed, the key
    :param default: mixed, default
    :return: mixed, default/actual
    """
    if key in self:
      # return stored if key exists
      return self.__getitem__(key)
    else:
      # set and return default if not
      self.__setitem__(key, default)
      return default

  def update(self, item = None, **kwargs):
    """
    Update self from dict/iterable

    Compatible with dict's update

    :param item: the dict/iterable
    :param kwargs: keyword arguments
    :return: None
    """
    # check if item is present
    if item is not None:
      # check if item has keys attribute
      if hasattr(item, 'keys'):
        for key in item.keys():
          self.__setitem__(key, item[key])
      else:
        # assumes item to be pairs
        for key, value in item:
          self.__setitem__(key, value)
    # always update self with kwargs
    for key, value in kwargs.items():
      self.__setitem__(key, value)

  def values(self):
    """
    Acquire all values in dictionary
    :return: generator, the values
    """
    yield from map(
      lambda _: self._loads(_),
      self._redis.hvals(self._token)
    )



# Author: Hansheng Zhao <copyrighthero@gmail.com> (https://www.zhs.me)

from .BaseStructure import BaseStructure

try:
  # attempts to import redis error
  from redis import ResponseError
except ModuleNotFoundError:
  # use general exception instead
  ResponseError = Exception


__all__ = ('HyperLogLog', )


class HyperLogLog(BaseStructure):

  __slots__ = ()

  def __init__(self, redis, token = None):
    """
    HyperLogLog interface constructor
    :param redis: redis, the instance
    :param token: mixed, access token
    """
    super(HyperLogLog, self).__init__(redis, token)
    # set the default data type
    self._type = b'string'
    self._initiate()
    # check if HLL string is valid
    try:
      self._redis.pfcount(self._token)
    except ResponseError:
      self._redis.delete(self._token)

  def __iadd__(self, value):
    """
    Augmented addition operation
    :param value: mixed, the value
    :return: self, the instance
    """
    self.register(self._dumps(value))
    return self

  def register(self, *args):
    """
    Register a new value into log
    :param args: mixed, the value(s)
    :return: None
    """
    self._redis.pfadd(
      self._token, *map(lambda _: self._dumps(_), args)
    )

  # aliases for register
  __add__ = __radd__ = register

  def cardinal(self):
    """
    Acquire count from log
    :return: int, the count
    """
    return self._redis.pfcount(self._token)

  # aliases for cardinal
  count = cardinal

  # log property for add and count
  log = property(cardinal, register)



# Author: Hansheng Zhao <copyrighthero@gmail.com> (https://www.zhs.me)

from .Set import Set
from .List import List
from .Dict import Dict
from .Queue import Queue
from .Serialize import Serialize
from .HyperLogLog import HyperLogLog


__all__ = (
  'Set', 'List', 'Dict', 'Queue', 'HyperLogLog',
  'Serialize', '__author__', '__version__', '__license__'
)

__author__ = 'Hansheng Zhao'
__license__ = 'BSD-2-Clause + MIT'
__version__ = '0!0.0b8'



# Author: Hansheng Zhao <zhaohans@msu.edu> (https://www.zhs.me)

from os import urandom
from .Serialize import Serialize


class BaseStructure(object):
  """ Redis interfaces base structure """

  __slots__ = (
    '_type', '_redis', '_token', '_serialize'
  )

  TOKEN_LENGTH = 16

  def __init__(self, redis, token = None):
    """
    BaseStructure class constructor
    :param redis: redis, the redis instance
    :param token: mixed, the access token
    """
    # set the default type
    self._type = b'none'
    # preserve redis instance
    self._redis = redis
    # assign remote access token
    self._token = urandom(self.TOKEN_LENGTH) \
      if token is None else self._convert(token)
    # set default serializer
    self._serialize = Serialize(
      serialize = 'msgpack', compress = 'zlib'
    )

  @property
  def token(self):
    """
    Token property getter
    :return: bytes, the token
    """
    return self._token

  @property
  def type(self):
    """
    Acquire token's redis type
    :return: bytes, the data type
    """
    return self._redis.type(self._token)

  @property
  def serialize(self):
    """
    Acquire instance serializer
    :return: instance, serializer
    """
    return self._serialize

  @serialize.setter
  def serialize(self, serialize):
    """
    Set instance serializer
    :param serialize: serializer
    :return: None
    """
    self._serialize = serialize

  @staticmethod
  def _convert(key):
    """
    Convert key to supported format
    :param key: mixed, any hashable object
    :return: bytes, bytes representation of key
    """
    # check if key is supported or hashable
    # support bytes by default
    if isinstance(key, bytes):
      return key
    # encode string type to bytes
    elif isinstance(key, str):
      return key.encode(encoding = 'UTF8')
    # change byte-array into bytes
    elif isinstance(key, bytearray):
      return bytes(key)
    # convert numeric into string
    elif isinstance(key, (int, float, complex)):
      return str(key).encode(encoding = 'UTF8')
    # convert hashable collections into string
    elif isinstance(key, (range, tuple, frozenset)):
      return str(key).encode(encoding = 'UTF8')
    # other types not supported
    else:
      raise TypeError('Unsupported type.')

  def _initiate(self):
    """
    Initiate the redis storage
    :return: None
    """
    # check if token exists in redis server
    if self.type not in (b'none', self._type):
      # remove token of different types
      self._redis.delete(self._token)

  def _dumps(self, payload):
    """
    Serialize value using serializer
    :param payload: mixed, value
    :return: bytes, the result
    """
    return self._serialize.dumps(payload)

  def _loads(self, payload):
    """
    Un-serialize value using serializer
    :param payload: bytes|string, payload
    :return: mixed, the object
    """
    return self._serialize.loads(payload)

  def delete(self):
    """
    Delete token from redis
    :return: bool, successful or not
    """
    return self._redis.delete(self._token) != 0

  # aliases for delete
  clear = flush = delete



# Author: Hansheng Zhao <copyrighthero@gmail.com> (https://www.zhs.me)

from .BaseStructure import BaseStructure


__all__ = ('Set', )


class Set(BaseStructure):

  __slots__ = ()

  def __init__(self, redis, token = None):
    """
    Set interface constructor
    :param redis: redis, the instance
    :param token: mixed, access token
    """
    super(Set, self).__init__(redis, token)
    # set the default data type
    self._type = b'set'
    self._initiate()

  def __contains__(self, item):
    """
    Check if item exists in set
    :param item: mixed, item
    :return: bool, whether exists
    """
    return self._redis.sismember(
      self._token, self._dumps(item)
    )

  def __eq__(self, other):
    """
    Check if two sets are the same
    :param other: mixed, sequence
    :return: bool, whether equals
    """
    if isinstance(other, Set):
      return self._content == other._content \
        if self.length == other.length else False
    else:
      return self.content == other

  def __gt__(self, other):
    """
    Check if is strict superset
    :param other: mixed, sequence
    :return: bool, whether true
    """
    return self.__ne__(other) and self.__ge__(other)

  def __iter__(self):
    """
    Content iterator
    :return: generator
    """
    yield from map(
      lambda _: self._loads(_),
      self._redis.smembers(self._token)
    )

  def __len__(self):
    """
    Acquire current set length
    :return: int, the set length
    """
    return self._redis.scard(self._token)

  def __lt__(self, other):
    """
    Check if is strict subset of other
    :param other: mixed, sequence
    :return: bool, whether true
    """
    return self.__ne__(other) and self.__le__(other)

  def __ne__(self, other):
    """
    Check if not equal to other
    :param other: mixed, sequence
    :return: bool, whether true
    """
    return not self.__eq__(other)

  def __hash__(self):
    """
    Return the hash of the content
    :return: int, the hash
    """
    return self._content.__hash__()

  @property
  def _content(self):
    """
    Acquire raw contents in set
    :return: set, the raw content
    """
    return set(self._redis.smembers(self._token))

  @property
  def content(self):
    """
    Acquire parsed contents
    :return: set, the parsed content
    """
    return set(map(
      lambda _: self._loads(_),
      self._redis.smembers(self._token)
    ))

  @property
  def length(self):
    """
    Acquire current set length
    :return: int, set length
    """
    return self.__len__()

  @staticmethod
  def _filter(iterable):
    """
    Filter out tokens and other instances
    :param iterable: tuple, the arguments
    :return: tuple<map, map>, the results
    """
    # filter out tokens and other instances
    return (map(lambda _: _.token, filter(
      lambda _: isinstance(_, Set), iterable
    )), filter(
      lambda _: not isinstance(_, Set), iterable
    ))

  def add(self, *args):
    """
    Add a member to the set
    :param args: mixed, stray args
    :return: int, unique item added
    """
    return self._redis.sadd(
      self._token,
      *map(lambda _: self._dumps(_), args)
    )

  def copy(self):
    """
    Create a new instance
    :return: instance
    """
    instance = Set(self._redis, self._token)
    instance.serialize = self._serialize
    return instance

  def difference(self, *args):
    """
    Get the difference between sequences
    :param args: other sequences
    :return: set, the difference
    """
    tokens, others = self._filter(args)
    # return the differences
    return set(map(
      lambda _: self._loads(_),
      self._redis.sdiff(self._token, *tokens)
    )).difference(*others)

  # aliases for differences
  __sub__ = __rsub__ = difference

  def difference_update(self, *args):
    """
    Acquire difference and update self
    :param args: other sequences
    :return: None
    """
    tokens, others = self._filter(args)
    # update differences using tokens
    self._redis.sdiffstore(
      self._token, self._token, *tokens
    )
    # discard any common items from set
    self.discard(
      *self.content.intersection(*others)
    )

  # aliases for difference update
  __isub__ = difference_update

  def discard(self, *args):
    """
    Discard members from set
    :param args: mixed, stray values
    :return: int, removed count
    """
    return self._redis.srem(
      self._token,
      *map(lambda _: self._dumps(_), args)
    )

  # aliases for discard
  remove = discard

  def intersection(self, *args):
    """
    Get intersection of sequences
    :param args: other sequences
    :return: set, the intersection
    """
    tokens, others = self._filter(args)
    # redis intersect using tokens
    intersect = set(map(
      lambda _: self._loads(_),
      self._redis.sinter(self._token, *tokens)
    ))
    # return the intersection of all items
    return intersect.intersection(*others)

  # aliases for intersection
  __and__ = __rand__ = intersection

  def intersection_update(self, *args):
    """
    Get intersection and update self
    :param args: other sequences
    :return: None
    """
    tokens, others = self._filter(args)
    # update intersections using tokens
    self._redis.sinterstore(
      self._token, self._token, *tokens
    )
    # discard any stray items from set
    self.discard(
      *self.content.difference(*others)
    )

  # aliases for intersection update
  __iand__ = intersection_update

  def isdisjoint(self, other):
    """
    Check if self and other is disjoint
    :param other: other sequence
    :return: bool, whether is disjoint
    """
    if isinstance(other, Set):
      return len(self._redis.sinter(
        self._token, other._token
      )) == 0
    else:
      return self.content.isdisjoint(other)

  def issubset(self, other):
    """
    Check if self is a subset of other
    :param other: mixed, sequence
    :return: bool, whether is subset
    """
    if isinstance(other, Set):
      return self._redis.sunion(
        self._token, other._token
      ) == other._content
    else:
      return self.content.issubset(other)

  # aliases for issubset
  __le__ = issubset

  def issuperset(self, other):
    """
    Check if self is a superset of other
    :param other: mixed, sequence
    :return: bool, whether is superset
    """
    if isinstance(other, Set):
      return self._redis.sunion(
        self._token, other._token
      ) == self._content
    else:
      return self.content.issuperset(other)

  # aliases for issuperset
  __ge__ = issubset

  def pop(self):
    """
    Return an arbitrary element
    :return: mixed, an item
    """
    return self._loads(self._redis.spop(self._token))

  def symmetric_difference(self, other):
    """
    Get symmetric diff of sequences
    :param other: mixed, sequence
    :return: set, the intersection
    """
    return self.union(other) - self.intersection(other)

  # aliases for symmetric difference
  __xor__ = __rxor__ = symmetric_difference

  def symmetric_difference_update(self, other):
    """
    Get symmetric diff and update self
    :param other: mixed, sequence
    :return: None
    """
    intersect = self.intersection(other)
    self.union_update(other)
    self.discard(*intersect)

  # aliases for symmetric difference update
  __ixor__ = symmetric_difference_update

  def union(self, *args):
    """
    Get the union of sequences
    :param args: other sequences
    :return: set, the intersection
    """
    tokens, others = self._filter(args)
    # redis union using tokens
    union = set(map(
      lambda _: self._loads(_),
      self._redis.sunion(self._token, *tokens)
    ))
    # return the intersection of all items
    return union.union(*others)

  # aliases for union
  __or__ = __ror__ = union

  def union_update(self, *args):
    """
    Get the union and update self
    :param args: other sequences
    :return: None
    """
    tokens, others = self._filter(args)
    # update union using tokens
    self._redis.sunionstore(
      self._token, self._token, *tokens
    )
    # add any stray items to set
    self.update(*others)

  # aliases for union update
  __ior__ = union_update

  def update(self, *args):
    """
    Update the set with members
    :param args: other iterable
    :return: None
    """
    for item in args: self.add(*item)



# Author: Hansheng Zhao <copyrighthero@gmail.com> (https://www.zhs.me)

# import required setup libraries
from setuptools import setup, find_packages
from codecs import open
from os import path

# import library for metadata
import redistr

# project absolute directory
DIRECTORY = path.abspath(path.dirname(__file__))

# project readme file content
with open(
  path.join(DIRECTORY, 'README.rst'), encoding = 'UTF8'
) as file_descriptor:
  PROJECT_README = file_descriptor.read()

# project required dependencies
with open(
  path.join(DIRECTORY, 'requirements.txt'), encoding = 'UTF8'
) as file_descriptor:
  REQUIREMENTS = tuple(line for line in file_descriptor if line)

# project setup parameters
setup(
  name = 'Redistr',
  version = redistr.__version__,
  description = 'Redis backed Python data structure interface.',
  long_description = PROJECT_README,
  url = 'https://www.github.com/copyrighthero/Redistr',
  download_url = 'https://www.github.com/copyrighthero/Redistr',
  author = redistr.__author__,
  author_email = 'copyrighthero@gmail.com',
  license = redistr.__license__,
  classifiers = (
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: BSD License',
    'License :: OSI Approved :: MIT License',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3 :: Only'
  ),
  keywords = 'Redis Remote Data-structure Multi-processing',
  # py_modules = ("Redistr", ),
  packages = find_packages(exclude = ()),
  install_requires = REQUIREMENTS,
  package_data = {},
  data_files = (),
  entry_points = {},
  project_urls = {
    'Source': 'https://www.github.com/copyrighthero/Redistr'
  }
)



# Author: Hansheng Zhao <copyrighthero@gmail.com> (https://www.zhs.me)

from .BaseStructure import BaseStructure


__all__ = ('List', )


class List(BaseStructure):
  """ Redis list structure interface """

  __slots__ = ()

  def __init__(self, redis, token = None):
    """
    List structure interface constructor
    :param redis: redis, the redis instance
    :param token: mixed, the access token
    """
    # invoke parent constructor
    super(List, self).__init__(redis, token)
    # set the default data type
    self._type = b'list'
    self._initiate()

  def __add__(self, other):
    """
    Perform addition operation
    :param other: list|tuple|range, sequences
    :return: list, the added list
    """
    return self.content + (other if not isinstance(
      other, (list, tuple, range)
    ) else list(other))

  def __contains__(self, value):
    """
    Check if value exists in list

    O(N) operation, not encouraged

    :param value: mixed, value
    :return: bool, if value exists
    """
    # serialize value
    value = self._dumps(value)
    return value in self._content

  def __delitem__(self, index):
    """
    Delete value at index from redis

    O(N) operation, not encouraged

    :param index: int, index
    :return: None
    """
    # acquire resources
    redis = self._redis
    token = self._token
    transform = self._transform
    # acquire list lengths
    length = self.length
    half_length = length // 2
    # check if index provided
    if isinstance(index, int):
      # trim off the first element
      if index == 0: redis.ltrim(token, 1, -1)
      # trim off the last element
      elif index == -1: redis.ltrim(token, 0, -2)
      else:
        # transform negative index
        index = transform(index)
        # index is on the left side
        if 0 < index <= half_length:
          # acquire left side segment
          segment = redis.lrange(token, 0, index - 1)
          # trim to the rest of the list
          redis.ltrim(token, index + 1, -1)
          # prepend the left side segment
          redis.lpush(token, *reversed(segment))
        # index is on the right side
        elif half_length < index < length :
          # acquire right side segment
          segment = redis.lrange(token, index + 1, -1)
          # trim the rest of the list
          redis.ltrim(token, 0, index - 1)
          # append the right side segment
          redis.rpush(token, *segment)
    # check if slice provided
    elif isinstance(index, slice):
      # warning: slow operation
      if index.step:
        # acquire list of indices
        indices = tuple(range(
          index.start, index.stop, index.step
        ))
        # remove one by one
        while indices:
          # acquire index
          index = indices[0]
          # shift indices by one
          indices = tuple(map(
            lambda _: _ - 1, indices[1:]
          )) if indices else []
          # delete current item
          self.__delitem__(index)
      else:
        # transform negative index
        start = transform(index.start)
        stop = transform(index.stop)
        # trim off left
        if start == 0: redis.ltrim(token, stop, -1)
        # trim off right
        elif stop == length: redis.ltrim(token, 0, start - 1)
        # check if slice range is valid
        elif 0 < start <= stop <= length:
          # trim from the left side
          if start + 1 <= length - stop:
            segment = redis.lrange(token, 0, start - 1)
            redis.ltrim(token, stop, -1)
            redis.lpush(token, *reversed(segment))
          # trim from the right side
          else:
            segment = redis.lrange(token, stop, -1)
            redis.ltrim(token, 0, start - 1)
            redis.rpush(token, *segment)

  def __eq__(self, other):
    """
    Check if two sequences are the same
    :param other: sequence
    :return: bool, the same
    """
    if len(self) == len(other):
      return list(self) == (
        list(other)
        if isinstance(other, (List, tuple, range))
        else other
      )
    else:
      return False

  def __getitem__(self, index):
    """
    Get value from redis list
    :param index: int|slice, the index
    :return: mixed|list, the value(s)
    """
    # acquire resources
    redis = self._redis
    token = self._token
    # default result
    result = None
    # check if index provided
    if isinstance(index, int):
      # use index to acquire item
      result = redis.lindex(token, index)
    # check if slice provided
    elif isinstance(index, slice):
      # acquire a range of items
      result = redis.lrange(
        token, index.start, index.stop - 1
      )
      # TODO: IMPLEMENT RIGHT STEPPING
      # WARNING: INCOMPATIBLE STEPPING
      if index.step:
        result = result[::index.step]
    # deserialize result
    if isinstance(result, bytes):
      result = self._loads(result)
    elif isinstance(result, list):
      result = list(map(
        lambda _: self._loads(_), result
      ))
    return result

  def __ge__(self, other):
    """
    Check greater or equal to other
    :param other: sequence
    :return: bool, comparison
    """
    return not self.__lt__(other)

  def __gt__(self, other):
    """
    Check greater than other
    :param other: sequence
    :return: bool, comparison
    """
    return not (self.__lt__(other) or self.__eq__(other))

  def __iadd__(self, other):
    """
    Append items to the list
    :param other: list|tuple|range, sequence
    :return: self, this instance
    """
    if isinstance(other, (list, tuple, range)):
      self._redis.rpush(self._token, *map(
        lambda _: self._dumps(_), other
      ))
    return self

  def __imul__(self, other):
    """
    Append a multiplication of items
    :param other: int, the multiplier
    :return: self, this instance
    """
    # check if multiplier is valid
    if isinstance(other, int):
      # acquire resources
      redis = self._redis
      token = self._token
      # append new items or flush
      redis.rpush(
        token, *(self._content * (other - 1))
      ) if other >= 1 else self.clear()
    return self

  def __iter__(self):
    """
    Parsed content iterator
    :return: generator
    """
    yield from map(
      lambda _: self._loads(_),
      self._redis.lrange(self._token, 0, -1)
    )

  def __len__(self):
    """
    Acquire list length
    :return: int, the length
    """
    return self._redis.llen(self._token)

  def __le__(self, other):
    """
    Check less than or equal to other
    :param other: sequence
    :return: bool, comparison
    """
    return self.__lt__(other) or self.__eq__(other)

  def __lt__(self, other):
    """
    Check if less than other
    :param other: sequence
    :return: bool, comparison
    """
    if len(self) == len(other):
      pass
    elif isinstance(other, (List, list, tuple, range)):
      return len(self) < len(other)
    else:
      raise TypeError('Invalid comparison types.')

  def __mul__(self, other):
    """
    Perform multiplication operation
    :param other: int, the multiplier
    :return: list, the result list
    """
    return self.content * other

  # alias for multiply
  __rmul__ = __mul__

  def __ne__(self, other):
    """
    Check if not equal to other
    :param other: sequence
    :return: bool, comparison
    """
    return not self.__eq__(other)

  def __radd__(self, other):
    """
    Perform addition operation
    :param other: list|tuple|range, sequences
    :return: list, the added list
    """
    return (other if not isinstance(
      other, (list, tuple, range)
    ) else list(other)) + self.content

  def __reversed__(self):
    """
    Acquire reversed generator
    :return: generator, the reversed
    """
    yield from reversed(self.content)

  def __setitem__(self, index, value):
    """
    Set value to index in redis list
    :param index: int, index
    :param value: mixed, value
    :return: None
    """
    # serialize value
    value = self._dumps(value)
    self._redis.lset(self._token, index, value)

  @property
  def _content(self):
    """
    Acquire raw list content
    :return: list, the content
    """
    return self._redis.lrange(self._token, 0, -1)

  @property
  def length(self):
    """
    Length property getter
    :return: int, the length
    """
    return self.__len__()

  @property
  def content(self):
    """
    Acquire deserialized list content
    :return: list, the content
    """
    return list(map(
      lambda _: self._loads(_), self._content
    ))

  def _transform(self, index):
    """
    Transform negative index
    :param index: int, index
    :return: int, index
    """
    # check if index is integer
    if isinstance(index, int):
      # acquire length
      length = self.length
      # return positive if valid
      return length + index \
        if -length < index < 0 else index
    else:
      # do nothing
      return index

  def append(self, value):
    """
    Append value to the end
    :param value: mixed, value
    :return: None
    """
    # serialize value
    value = self._dumps(value)
    self._redis.rpush(self._token, value)

  push = append

  def circulate(self, token = None):
    """
    Circulate an item (RPOPLPUSH)
    :param token: None|mixed, the target
    :return: mixed, the item
    """
    # convert token to bytes
    token = self._convert(token) \
      if token is not None else self._token
    # acquire and circulate an item
    result = self._redis.rpoplpush(self._token, token)
    # deserialize result
    if result is not None: result = self._loads(result)
    return result

  def copy(self):
    """
    Acquire a new instance
    :return: instance, the instance
    """
    instance = List(self._redis, self._token)
    instance.serialize = self.serialize
    return instance

  def count(self, value):
    """
    Count the value in list
    :return: int, the count
    """
    # serialize value
    value = self._dumps(value)
    return self._content.count(value)

  def extend(self, iterable):
    """
    Extend list with iterable
    :param iterable: iterable
    :return: None
    """
    self._redis.rpush(self._token, *map(
      lambda _: self._dumps(_), iterable
    ))

  def index(self, value, start = None, stop = None):
    """
    Acquire value's index
    :param value: mixed, the value
    :param start: the start index
    :param stop: the stop index
    :return: int, the index
    """
    # serialize value
    value = self._dumps(value)
    return self._content.index(value, start, stop)

  def insert(self, index, value):
    """
    Insert an item to index
    :param index: int, the index
    :param value: mixed, the value
    :return: None
    """
    # serialize value
    value = self._dumps(value)
    self._redis.lset(self._token, index, value)

  def pop(self, index = None, flag = True):
    """
    Return and delete an item
    :param index: int|None, the index
    :param flag: bool, from right or not
    :return: mixed|None, the item
    """
    # acquire resources
    redis = self._redis
    token = self._token
    # acquire list length
    length = self.length
    # default result
    result = None
    # check if index provided
    if index is None:
      # regular pop from list
      result = redis.rpop(token) \
        if flag else redis.lpop(token)
    # check if index is integer
    elif isinstance(index, int):
      # transform negative index
      index = self._transform(index)
      # regular left pop operation
      if index == 0:
        result = redis.lpop(token)
      # regular right pop operation
      elif index == length - 1:
        result = redis.rpop(token)
      elif 0 < index < length - 1:
        # acquire and remove item
        result = self.__getitem__(index)
        self.__delitem__(index)
    # check if slice provided
    elif isinstance(index, slice):
      # acquire and remove item
      result = self.__getitem__(index)
      self.__delitem__(index)
    # deserialize result
    if isinstance(result, bytes):
      result = self._loads(result)
    elif isinstance(result, list):
      result = list(map(
        lambda _: self._loads(_), result
      ))
    return result

  def popleft(self):
    """
    Pop an item from left
    :return: mixed|None, the item
    """
    result = self._redis.lpop(self._token)
    return self._loads(result) if result else result

  # aliases for popleft
  shift = popleft

  def prepend(self, value):
    """
    Prepend value to the start
    :param value: mixed, value
    :return: None
    """
    # serialize value
    value = self._dumps(value)
    self._redis.lpush(self._token, value)

  # aliases for prepend
  unshift = prepend

  def remove(self, value):
    """
    Remove all value from list
    :param value: mixed, value
    :return: int, actual count
    """
    # serialize value
    value = self._dumps(value)
    return self._redis.lrem(self._token, value)

  # aliases for remove
  discard = purge = remove

  def reverse(self):
    """
    Reverse the list from redis
    :return: None
    """
    # acquire raw content
    content = self._content
    # clear and reset list
    self.clear()
    self._redis.rpush(
      self._token, *reversed(content)
    )

  def sort(self, *args, **kwargs):
    """
    Sort the list from redis
    :param args: the arguments
    :param kwargs: keyword args
    :return: None
    """
    # acquire and sort content
    content = self.content
    content.sort(*args, **kwargs)
    # clear and reset list
    self.clear()
    self._redis.rpush(self._token, *map(
      lambda _: self._dumps(_), content
    ))
